# -*- coding: utf-8 -*-
__author__ = 'ikibalin'
__version__ = "2020_08_21"

import os
import os.path
import sys

from typing import NoReturn
from PyQt5 import QtCore, QtGui, QtWidgets

import json

from cryspy import GlobalN, file_to_globaln, RhoChi

from cryspy_editor.widgets.w_panel import WPanel, \
    find_tree_item_position, L_GLOBAL_CLS
from cryspy_editor.widgets.w_methods import WMethods
from cryspy_editor.widgets.w_baseobj import WBaseObj
from cryspy_editor.widgets.w_dockarea import WDockArea
from cryspy_editor.widgets.w_console import WConsole
from cryspy_editor.widgets.w_function import WFunction

# import importlib
# dir_additional_libraries = d_setup['directory_additional_libraries']


def text_from_d_info(d_info: dict = None):
    s_out = ""
    if d_info is not None:
        s_out = d_info["print"]
        if "d_info" in d_info.keys():
            s_out_2 = text_from_d_info(d_info["d_info"])
            s_out += "\n\n" + s_out_2
    return s_out


class CCalcRun(QtWidgets.QMainWindow):
    """During calculations."""

    def __init__(self, parent=None):
        super(CCalcRun, self).__init__(parent)
        self.d_info = None
        self.init_widget()
        self.setWindowTitle(
            'Calculations are running ... (DO NOT CLOSE THE WINDOW)')

    def init_widget(self):
        """Init widget."""
        cw = QtWidgets.QWidget(self)
        self.layout = QtWidgets.QHBoxLayout()
        lay_v = QtWidgets.QVBoxLayout()

        self.label = WConsole(cw)
        self.btn = QtWidgets.QPushButton()
        self.btn.clicked.connect(self.stop_calc)
        lay_v.addWidget(self.label)
        lay_v.addWidget(self.btn)

        self.layout.addLayout(lay_v)
        cw.setLayout(self.layout)
        self.setCentralWidget(cw)
        self.timer = QtCore.QBasicTimer()

    def timerEvent(self, e):
        s_text = text_from_d_info(self.d_info)
        self.label.setText(s_text)

    def set_d_info(self, d_info: dict = None):
        self.d_info = d_info
        if self.d_info is not None:
            self.btn.setEnabled(True)
            self.btn.setText("Stop")
            self.label.setText(self.d_info["print"])

    def stop_calc(self):
        if self.d_info is not None:
            self.d_info["stop"] = True
            self.btn.setText("Wait a little.")
            self.btn.setEnabled(False)


class MyThread(QtCore.QThread):
    """Thread class."""

    signal_start = QtCore.pyqtSignal()
    signal_end = QtCore.pyqtSignal()
    signal_refresh = QtCore.pyqtSignal()
    signal_take_attributes = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        QtCore.QThread.__init__(self, parent)
        self.message = None
        self.function = None
        self.arguments = None
        self.output = None
        self.d_info = None

    def run(self):
        """Run."""
        func = self.function
        arg = self.arguments
        n_row_need = func.__code__.co_argcount
        l_var_name = func.__code__.co_varnames[:n_row_need]

        if "d_info" in l_var_name:
            self.d_info = {"stop": False, "print": ""}
            self.window.set_d_info(self.d_info)
            flag_d_info = True
        else:
            flag_d_info = False
            self.d_info = None

        self.signal_start.emit()

        try:
            if flag_d_info:
                out = func(*arg, d_info=self.d_info)
            else:
                out = func(*arg)
        except Exception as e:
            print(e)
            out = "MISTAKE DURING PROGRAM EXECUTION\n\n"+str(e)
        self.output = out
        self.signal_end.emit()


class CBuilder(QtWidgets.QMainWindow):
    """
    Builder class.

    Attributes
    ----------
        - f_dir_prog
        - f_dir_data
        - f_file
        - thread
        - wfunction
        - wpanel
        - wconsole
        - object
        - wbaseobj
        - wdockarea
        - wcb_for_stackedwidget
        - wstackwidget
    """

    def __init__(self, f_dir_data=os.path.dirname(__file__)):
        super(CBuilder, self).__init__()

        self.f_dir_prog = os.path.dirname(__file__)
        self.index_curent_item = None

        self.def_actions()
        self.init_widget()

        self.mythread = MyThread(self)
        self.mythread.signal_start.connect(self.start_of_calc_in_thread)
        self.mythread.signal_end.connect(self.end_of_calc_in_thread)
        self.mythread.signal_refresh.connect(self.form_wpanel)

        self.mythread.window = CCalcRun(self)
        # self.mythread.window.setWindowFlags(QtCore.Qt.CustomizeWindowHint |
        #                                     QtCore.Qt.WindowCloseButtonHint)

        screen = QtWidgets.QDesktopWidget().screenGeometry()
        self.mythread.window.resize(screen.width() * 0.5,
                                    screen.height() * 0.2)
        f_calc = os.path.join(self.f_dir_prog, "f_icon", "calc.gif")
        q_label = QtWidgets.QLabel()
        gif = QtGui.QMovie(f_calc)
        q_label.setMovie(gif)
        gif.start()
        self.mythread.window.layout.addWidget(q_label)

        self.wpanel.set_func_object_clicked(self.onItemClicked)

        self.wmethods.set_wfunction(self.wfunction)
        self.wmethods.set_thread(self.mythread)

        self.wdockarea.set_thread(self.mythread)
        self.wbaseobj.set_thread(self.mythread)

        self.f_dir_data = f_dir_data
        self.f_file = None
        self.wpanel.object = None
        if os.path.isfile(f_dir_data):
            self.f_file = f_dir_data
            self.f_dir_data = os.path.dirname(f_dir_data)
        os.chdir(self.f_dir_data)

        self.setWindowTitle('CrysPy editor')

        if self.f_file is not None:
            self.setWindowTitle(f'CrysPy editor: {self.f_file:}')
            self.load_globaln()
        else:
            # self.wpanel.object = GlobalN.make_container((), (), "global")
            self.wpanel.object = RhoChi()
            self.form_wpanel()
            self.form_wstackedwidget(self.wpanel.object)

        self.show()

    def def_actions(self):
        """Define actions."""
        f_dir_prog = self.f_dir_prog
        f_dir_prog_icon = os.path.join(f_dir_prog, 'f_icon')
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(f_dir_prog_icon, 'icon.png')))

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        self.toolbar = self.addToolBar("Open")

        new_action = QtWidgets.QAction(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'new.png')), '&New', self)
        new_action.setStatusTip('New')
        new_action.triggered.connect(self.new)
        fileMenu.addAction(new_action)
        self.toolbar.addAction(new_action)

        open_action = QtWidgets.QAction(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'open.png')), '&Open', self)
        open_action.setShortcut('Ctrl+O')
        open_action.setStatusTip('Open file')
        open_action.triggered.connect(self.open)
        fileMenu.addAction(open_action)
        self.toolbar.addAction(open_action)

        save_action = QtWidgets.QAction(QtGui.QIcon(
            os.path.join(f_dir_prog_icon, 'save.png')), '&Save', self)
        save_action.setShortcut('Ctrl+S')
        save_action.setStatusTip('Save')
        save_action.triggered.connect(self.save)

        fileMenu.addAction(save_action)
        self.toolbar.addAction(save_action)

        save_as_action = QtWidgets.QAction(
            QtGui.QIcon(os.path.join(f_dir_prog_icon, 'save_as.png')),
            'Save &as...', self)
        save_as_action.setStatusTip('Save as ...')
        save_as_action.triggered.connect(self.save_as)

        fileMenu.addAction(save_as_action)
        self.toolbar.addAction(save_as_action)

        # refresh_wind = QtWidgets.QAction(
        #     QtGui.QIcon(os.path.join(f_dir_prog_icon, 'refresh.png')),
        #     'Refresh', self)
        # refresh_wind.setStatusTip('Refresh')
        # refresh_wind.triggered.connect(self.form_wpanel)
        # self.toolbar.addAction(refresh_wind)

        open_folder = QtWidgets.QAction(
            QtGui.QIcon(os.path.join(f_dir_prog_icon, 'open_folder.png')),
            'Open folder', self)
        open_folder.setStatusTip('Open folder')
        open_folder.triggered.connect(self.open_folder)
        self.toolbar.addAction(open_folder)

        # flip_action = QtWidgets.QAction(
        #     QtGui.QIcon(os.path.join(f_dir_prog_icon, 'flip.png')),
        #     'Flip representation', self)
        # flip_action.setStatusTip('Flip presentation')
        # flip_action.triggered.connect(self.flip_presenation)
        # self.toolbar.addAction(flip_action)

        fileMenu = menubar.addMenu('&Object manipulations')
        manip_action = QtWidgets.QAction('Switch to global', fileMenu)
        manip_action.object = GlobalN
        manip_action.triggered.connect(self.transfer)
        fileMenu.addAction(manip_action)
        for item_cls in L_GLOBAL_CLS:
            manip_action = QtWidgets.QAction(
                f'Switch to {item_cls.PREFIX:}', fileMenu)
            manip_action.object = item_cls
            manip_action.triggered.connect(self.transfer)
            fileMenu.addAction(manip_action)

        self.statusBar()

    def transfer(self):
        """Transfer GlobalN object to any one."""
        sender = self.sender()
        cls_item = sender.object
        item_old = self.wpanel.object
        if cls_item is GlobalN:
            item_new = cls_item.make_container(
                (), tuple(set([type(item) for item in item_old])), "global")
        else:
            item_new = cls_item()
        item_new.add_items(item_old.items)
        self.wpanel.object = item_new
        self.form_wpanel()
        self.form_wstackedwidget(self.wpanel.object)

    def flip_presenation(self) -> NoReturn:
        """Flip presentation."""
        ind = 1-self.wstackedwidget.currentIndex()
        self.wstackedwidget.setCurrentIndex(ind)

    def init_widget(self) -> NoReturn:
        """Initilize widget."""
        self.location_on_the_screen()

        widget_main = QtWidgets.QWidget(self)
        layout_main = QtWidgets.QVBoxLayout()
        self.wfunction = WFunction(widget_main)

        # First
        layout_main.addWidget(self.wfunction)

        width_m_1 = self.info_width / 6.
        width_m_2 = (3 * self.info_width) / 6.
        width_v_1 = self.info_height * 0.75
        width_v_2 = self.info_height * 0.25

        # Second
        wsplitter = QtWidgets.QSplitter(widget_main)

        # Panel from left site
        self.wpanel = WPanel(wsplitter)
        wsplitter.addWidget(self.wpanel)

        # Central widget constracted from three widgets
        wsplitter_centr = QtWidgets.QSplitter(wsplitter)
        wsplitter_centr.setOrientation(QtCore.Qt.Vertical)

        # 1.
        self.wstackedwidget = QtWidgets.QStackedWidget(wsplitter_centr)
        self.wbaseobj = WBaseObj(self.wstackedwidget)
        self.wdockarea = WDockArea(self.wstackedwidget)
        self.wstackedwidget.addWidget(self.wbaseobj)
        self.wstackedwidget.addWidget(self.wdockarea)
        wsplitter_centr.addWidget(self.wstackedwidget)

        # 2.
        self.wconsole = WConsole(wsplitter_centr)
        wsplitter_centr.addWidget(self.wconsole)

        wsplitter_centr.setSizes([width_v_1, width_v_2])
        wsplitter.addWidget(wsplitter_centr)

        # Panel from right site
        self.wmethods = WMethods(wsplitter)
        wsplitter.addWidget(self.wmethods)

        wsplitter.setSizes([width_m_1, width_m_2, 0])

        layout_main.addWidget(wsplitter)
        widget_main.setLayout(layout_main)
        self.setCentralWidget(widget_main)

    def location_on_the_screen(self) -> NoReturn:
        """Location on the screen."""
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        self.setMinimumSize(screen.width() * 1 / 4, screen.height() * 1 / 4)
        self.info_width = screen.width() * 8. / 10.
        self.info_height = screen.height() * 14. / 16.
        self.move(screen.width() / 10, screen.height() / 20)
        self.resize(screen.width() * 8. / 10., screen.height() * 14. / 16.)


    def calc_is_finished(self) -> NoReturn:
        """After calculations."""
        m_box = QtWidgets.QMessageBox(self)
        m_box.setIcon(QtWidgets.QMessageBox.Information)
        m_box.setText("Calculations are finished")
        m_box.setWindowTitle("Message")
        m_box.setStandardButtons(
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        m_box.exec()

        # self.load_globaln()
    def load_globaln(self) -> NoReturn:
        """Read globaln from file."""
        obj = file_to_globaln(self.f_file)
        self.wpanel.object = obj
        self.form_wpanel()
        self.form_wstackedwidget(obj)
        self.wmethods.get_methods(obj)

        w_item = self.wpanel.topLevelItem(0)
        self.wpanel.setCurrentItem(w_item)

    def new(self) -> NoReturn:
        """Define new object."""
        obj = RhoChi()
        self.wpanel.object = obj
        self.form_wpanel()
        self.form_wstackedwidget(obj)
        self.wmethods.get_methods(obj)

    def save(self) -> NoReturn:
        """Save."""
        f_name = self.f_file
        obj = self.wpanel.object
        if type(obj) is RhoChi:
            obj.save_to_file(f_name)
        else:
            with open(f_name, "w") as fid:
                fid.write(str(obj))

    def save_as(self) -> NoReturn:
        """Save as."""
        f_name, okPressed = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Select a file:', self.f_dir_data, "Rcif files (*.rcif)")
        if not (okPressed):
            return None
        self.f_file = f_name
        self.setWindowTitle(f'CrysPy editor: {f_name:}')

        self.f_dir_data = os.path.dirname(f_name)
        os.chdir(os.path.dirname(f_name))
        self.save()

    def open_folder(self) -> NoReturn:
        """Open folder."""
        os.startfile(self.f_dir_data)

    def open(self) -> NoReturn:
        """Open."""
        f_name, okPressed = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Select a cif file:', self.f_dir_data, "All files (*.*)")
        if not (okPressed):
            return None
        self.f_file = f_name
        self.setWindowTitle(f'CrysPy editor: {f_name:}')
        self.f_dir_data = os.path.dirname(f_name)
        os.chdir(os.path.dirname(f_name))
        self.load_globaln()

        f_setup = os.path.join(os.path.dirname(__file__), 'setup.json')
        if os.path.isfile(f_setup):
            d_setup = json.load(open(f_setup, 'r'))
        else:
            d_setup = {}
        d_setup["last_directory"] = f_name
        s_cont = json.dumps(d_setup, sort_keys=True, indent=4)
        with open(f_setup, "w") as fid:
            fid.write(s_cont)

    # @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, int)
    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, QtWidgets.QTreeWidgetItem)
    def onItemClicked(self, it, col):
        """On item clicked."""
        if it is not None:
            self.form_wstackedwidget(it.object)

    def start_of_calc_in_thread(self):
        """Start calculations in thread."""
        cur_item = self.wpanel.currentItem()
        self.index_curent_item = tuple(find_tree_item_position(
            self.wpanel, cur_item))

        ls_out = ["The calculations are running.\n"]
        self.wconsole.setText("\n".join(ls_out))

        thread = self.mythread
        if thread.d_info is not None:
            thread.window.show()
            thread.window.timer.start(100, thread.window)
            
    # def timerEvent(self, a0):
    #     self.pbar.setValue(20)

    def end_of_calc_in_thread(self):
        """End calculations in thread."""
        
        self.wfunction.calculation_finished()

        thread = self.mythread
        thread.window.hide()
        thread.window.timer.stop()
 
        function = thread.function
        function_name = function.__name__
        output = thread.output
        ls_out = []
        ls_out.append(
            f"The function '{function_name:}' is perfomed.\n\nResult is \n")
        ls_out.append(str(output))
        self.wconsole.setText("\n".join(ls_out))

        self.form_wpanel()
        # self.form_wstackedwidget(self.wpanel.object)
        # self.wmethods.get_methods(self.wpanel.object)

        previous_item = self.index_curent_item

        if previous_item is not None:
            wpanel = self.wpanel
            n_main = previous_item[0]
            if n_main < wpanel.columnCount():
                w_item = wpanel.itemAt(n_main, 0)
                if len(previous_item) > 1:
                    for n_ind in previous_item[1:]:
                        if n_ind < w_item.childCount():
                            w_item = w_item.child(n_ind)
                        else:
                            break
        else:
            w_item = self.wpanel.topLevelItem(0)
        self.wpanel.setCurrentItem(w_item)
        self.form_wstackedwidget(w_item.object)
        self.wmethods.get_methods(w_item.object)

        # self.calc_is_finished()

    def form_wpanel(self):
        """Form wpanel."""
        self.wpanel.set_object_global()

    def form_wstackedwidget(self, obj):
        """From stackwidget."""
        self.wbaseobj.set_object(obj)
        self.wdockarea.set_object(obj)
        if self.wdockarea.count() == 0:
            self.wstackedwidget.setCurrentIndex(0)
        else:
            self.wstackedwidget.setCurrentIndex(1)
        self.wmethods.get_methods(obj)


def main_w(l_arg=[]):
    """Make main window."""
    app = QtWidgets.QApplication(l_arg)
    # app.setStyle(QtWidgets.QStyleFactory.create('Fusion'))
    app.setStartDragDistance(100)
    f_setup = os.path.join(os.path.dirname(__file__), 'setup.json')
    if len(l_arg) >= 2:
        f_dir_data = os.path.abspath(l_arg[1])
    elif os.path.isfile(f_setup):
        d_setup = json.load(open(f_setup, 'r'))
        f_dir_data = d_setup["last_directory"]
        if not(os.path.isdir(f_dir_data) | os.path.isfile(f_dir_data)):
            f_dir_data = os.getcwd()
            d_setup["last_directory"] = f_dir_data
            s_cont = json.dumps(d_setup, sort_keys=True, indent=4)
            with open(f_setup, "w") as fid:
                fid.write(s_cont)
    else:
        f_dir_data = os.getcwd()
    wobj = CBuilder(f_dir_data)
    sys.exit(app.exec_())


if __name__ == '__main__':
    l_arg = sys.argv
    main_w(l_arg)



"""WFunction class."""
from PyQt5 import QtWidgets, QtCore
from typing import NoReturn, Callable
import numpy


class WFunction(QtWidgets.QFrame):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WFunction, self).__init__(parent)
        self.setStyleSheet("background-color:white;")
        self.setFrameShape(QtWidgets.QFrame.Box)

        self.setSizePolicy(QtWidgets.QSizePolicy(
                    QtWidgets.QSizePolicy.Expanding,
                    QtWidgets.QSizePolicy.Fixed))

        self.layout_func = QtWidgets.QHBoxLayout()
        self.push_button = QtWidgets.QPushButton("Run function")
        self.w_cb_hide = QtWidgets.QCheckBox("hide")
        self.w_cb_hide.setCheckState(2)
        self.w_cb_hide.stateChanged.connect(self.show_hide)

        self.hide()

        layout_central = QtWidgets.QHBoxLayout()
        layout_central.addLayout(self.layout_func)
        layout_central.addStretch(1)
        layout_central.addWidget(self.push_button)
        layout_central.addWidget(self.w_cb_hide)

        self.setLayout(layout_central)

        self.l_w_arg = []
        self.function_attached = None
        self.thread_attached = None
        self.push_button.clicked.connect(self.run_function)

    def show_hide(self, *arg) -> NoReturn:
        """Show hide."""
        if arg[0] == 2:
            self.hide()
        else:
            self.show()

    def set_function(self, func: Callable, thread: QtCore.QThread):
        """Set function."""
        self.push_button.setText(f"Run '{func.__name__:}'")
        self.w_cb_hide.setCheckState(0)
        self.setToolTip(func.__doc__)
        n_row_need = func.__code__.co_argcount
        l_var_name = func.__code__.co_varnames[:n_row_need]

        if func.__defaults__ is None:
            l_defaults = []
        else:
            l_defaults = func.__defaults__
        n_defaults = len(l_defaults)
        n_var_names = len(l_var_name)

        layout = self.layout_func
        del_layout(layout)

        self.l_w_arg = []

        for _i_var, _var_name in enumerate(l_var_name):
            i_default = _i_var-(n_var_names-n_defaults)
            if i_default >= 0:
                s_def = str(l_defaults[i_default]) + " (default)"
            else:
                s_def = "<drop object>"
            if (_var_name != "self"):
                if _var_name != "d_info":
                    widget = DropLabel(f"{s_def:}")
                    widget.setStyleSheet(
                        "background:lightyellow; border: 2px solid red;")
                    if i_default >= 0:
                        widget.attached_object = l_defaults[i_default]
                        widget.setStyleSheet("")
                    layout.addWidget(QtWidgets.QLabel(_var_name))
                    layout.addWidget(widget)
                    self.l_w_arg.append(widget)

        self.function_attached = func
        self.thread_attached = thread

    def is_defined(self):
        """Is defined."""
        return all([w_arg.attached_object is not None
                    for w_arg in self.l_w_arg])

    def run_function(self) -> NoReturn:
        """Run function."""
        func = self.function_attached
        thread = self.thread_attached
        if func is None:
            return

        l_w_arg = self.l_w_arg
        self.push_button.setEnabled(False)
        l_x = [_.attached_object for _ in l_w_arg]
        t_x = tuple(l_x)
        if thread is None:
            try:
                func(*t_x)
            except Exception:
                pass
            self.calculation_finished()
        else:
            thread.function = func
            thread.arguments = t_x
            thread.start()

    def calculation_finished(self):
        """After calculations."""
        # self.w_cb_hide.setCheckState(2)
        self.push_button.setEnabled(True)


class DropLabel(QtWidgets.QLineEdit):  # FIXME: remove to another file
    """Drop label."""

    def __init__(self, *args, **kwargs):
        super(DropLabel, self).__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        self.attached_object = None
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.editingFinished.connect(self.convert_to_object)
        self.selectionChanged.connect(lambda: self.setText(""))

    def dragEnterEvent(self, event):
        """Drag enter event."""
        if event.mimeData().hasText():
            event.acceptProposedAction()

    def dropEvent(self, event):
        """Drop event."""
        # pos = event.pos()
        mime_data = event.mimeData()
        s_cont = mime_data.text()
        self.attached_object = mime_data.object_to_send
        self.setText(s_cont)
        self.setStyleSheet("")
        event.acceptProposedAction()

    def convert_to_object(self):
        """Convert to object."""
        text = self.text()
        obj = None
        flag = False
        flag_comma = (text.find(',') != -1)
        if flag_comma:
            l_text = text.split(",")
            try:
                obj = numpy.array(l_text, dtype=int)
                flag = True
            except Exception:
                pass
            if not flag:
                try:
                    obj = numpy.array(l_text, dtype=float)
                    flag = True
                except Exception:
                    pass
        else:
            try:
                obj = int(text)
                flag = True
            except Exception:
                pass
            if not flag:
                try:
                    obj = float(text)
                    flag = True
                except Exception:
                    pass
        if not flag:
            obj = str(text)
        self.attached_object = obj
        self.setAlignment(QtCore.Qt.AlignRight)
        self.setStyleSheet("")


def w_for_function(obj: Callable, layout_11: QtWidgets.QBoxLayout,
                    layout_12: QtWidgets.QBoxLayout,
                    layout_13: QtWidgets.QBoxLayout,
                    layout_2: QtWidgets.QBoxLayout,
                    layout_3: QtWidgets.QBoxLayout, w_output: QtWidgets.QWidget,
                    thread: QtCore.QThread) -> NoReturn:  # FIXME delete it
    """Widget for WFunction."""
    w_out = WFunction()
    w_out.set_function(obj, thread)
    layout_11.addWidget(w_out)


def del_layout(layout):  # FIXME delete it
    """Delete all elements from layouts."""
    for i in reversed(range(layout.count())):
        if layout.itemAt(i).widget() is not None:
            layout.itemAt(i).widget().setParent(None)
        elif layout.itemAt(i).layout() is not None:
            del_layout(layout.itemAt(i).layout())
            layout.itemAt(i).layout().setParent(None)
        else:
            layout.removeItem(layout.itemAt(i))



"""WPanel."""
from typing import Union, Callable, Any, NoReturn
import copy
from PyQt5 import QtCore, QtGui, QtWidgets

from cryspy import ItemN, LoopN, DataN, GlobalN, str_to_items, \
    str_to_globaln

from cryspy import L_GLOBAL_CLASS, L_DATA_CLASS, L_ITEM_CLASS, L_LOOP_CLASS
L_GLOBAL_CLS = L_GLOBAL_CLASS
L_DATA_CLS = L_DATA_CLASS
L_LOOP_CLS = L_LOOP_CLASS
L_ITEM_CLS = L_ITEM_CLASS
# import importlib
# additional_module = importlib.import_module("cryspy")


class WPanel(QtWidgets.QTreeWidget):
    """WPanel class."""

    def __init__(self, parent=None) -> NoReturn:
        super(WPanel, self).__init__(parent)
        self.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding))
        self.setColumnCount(1)
        self.setHeaderHidden(True)
        self.object = None

        self.customContextMenuRequested.connect(self.open_menu)
        # for key, item in kwargs.keys():
        #     if key == ""

    def set_object_global(self) \
            -> NoReturn:
        """Set object GlobalN."""
        obj_globaln = self.object
        if obj_globaln is None:
            obj_globaln = GlobalN.make_container((), (), "global")

        if self.topLevelItemCount() != 0:
            ind = 0
            w_del = self.takeTopLevelItem(ind)
            self.removeItemWidget(w_del, ind)

        wi = make_tree_widget_item(self, obj_globaln)
        self.addTopLevelItem(wi)

        self.addTopLevelItem(wi)
        self.expandAll()
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.setAcceptDrops(True)
        self.setDragEnabled(True)

    def set_func_object_clicked(self, func_object_clicked:
                                Callable[[Any], Any]) -> NoReturn:
        """Set function when object is clicked."""
        self.func_object_clicked = func_object_clicked
        # self.itemClicked.connect(func_object_clicked)
        self.currentItemChanged.connect(func_object_clicked)

    def dragEnterEvent(self, event):
        """Drag enter event."""
        if event.mimeData().hasText():
            event.acceptProposedAction()

    def dropEvent(self, event):  # FIXME It doesnot work, why?
        """Drop event."""
        print("HERE2_ DROP EVENT!!!!")
        pos = event.pos()
        mime_data = event.mimeData()
        s_cont = mime_data.text()
        object = mime_data.object_to_send
        w_given = self.itemAt(pos)
        print("s_cont:", s_cont)
        print("object:", object)
        print("w_given:", w_given)
        event.acceptProposedAction()

    def mouseMoveEvent(self, event):
        """Mouse move event."""
        w_tree_item = self.itemAt(event.pos())
        if w_tree_item is None:
            return
        object_to_send = w_tree_item.object
        string_to_send = type(object_to_send).__name__

        drag = QtGui.QDrag(self)
        dragMimeData = QtCore.QMimeData()
        dragMimeData.object_to_send = object_to_send

        dragMimeData.setText(string_to_send)
        drag.setMimeData(dragMimeData)
        drag.exec_(QtCore.Qt.MoveAction)

    def open_menu(self, position):
        """Context menu."""
        w_item = self.itemAt(position)
        if w_item is not None:
            # l_ind = find_tree_item_position(self, w_item)
            obj = w_item.object
            menu = QtWidgets.QMenu(self)
            save_as = QtWidgets.QAction('Save to file ...', menu)
            save_as.triggered.connect(lambda x: self.save_object(obj))
            menu.addAction(save_as)

            if isinstance(obj, (GlobalN, DataN)):
                add_items = QtWidgets.QAction('Add items by cif ...', menu)
                add_items.triggered.connect(lambda: self.add_items(w_item))
                menu.addAction(add_items)

            if ((type(obj) is GlobalN) | (type(obj) is DataN)):
                menu_item = menu.addMenu("Add")
                if type(obj) is GlobalN:
                    l_list = L_DATA_CLS + L_LOOP_CLS + L_ITEM_CLS
                else:
                    l_list = L_LOOP_CLS + L_ITEM_CLS

                for cls_item in l_list:
                    if "PREFIX" in cls_item.__dict__.keys():
                        prefix = cls_item.PREFIX
                    else:
                        prefix = cls_item.ITEM_CLASS.PREFIX
                    add_item = QtWidgets.QAction(f'{prefix :}',
                                                 menu_item)
                    add_item.cls_item = cls_item
                    add_item.triggered.connect(lambda: self.add_item(w_item))
                    menu_item.addAction(add_item)

            elif isinstance(obj, (GlobalN, DataN)):
                menu_item = menu.addMenu("Add")
                for cls_item in obj.CLASSES_MANDATORY:
                    if ((cls_item is not DataN) & (cls_item is not LoopN) &
                            (cls_item is not ItemN)):
                        if "PREFIX" in cls_item.__dict__.keys():
                            prefix = cls_item.PREFIX
                        else:
                            prefix = cls_item.ITEM_CLASS.PREFIX
                        add_item = QtWidgets.QAction(f'{prefix:}',
                                                     menu_item)
                        add_item.cls_item = cls_item
                        add_item.triggered.connect(
                            lambda: self.add_item(w_item))
                        menu_item.addAction(add_item)
                menu_item.addSeparator()
                for cls_item in obj.CLASSES_OPTIONAL:
                    if ((cls_item is not DataN) & (cls_item is not LoopN) &
                            (cls_item is not ItemN)):
                        if "PREFIX" in cls_item.__dict__.keys():
                            prefix = cls_item.PREFIX
                        else:
                            prefix = cls_item.ITEM_CLASS.PREFIX
                        add_item = QtWidgets.QAction(f'{prefix :}',
                                                     menu_item)
                        add_item.cls_item = cls_item
                        add_item.triggered.connect(
                            lambda: self.add_item(w_item))
                        menu_item.addAction(add_item)

            if isinstance(obj, (GlobalN, DataN, LoopN)):
                act_rename = QtWidgets.QAction('Rename', menu)
                act_rename.triggered.connect(lambda: self.rename_item(w_item))
                menu.addAction(act_rename)

            if isinstance(obj, (DataN, LoopN, ItemN)):
                del_item = QtWidgets.QAction('Delete', menu)
                del_item.triggered.connect(lambda: self.delete_object(w_item))
                menu.addAction(del_item)

            if isinstance(obj, GlobalN):
                menu_tr = menu.addMenu("Transfer")
                qaction = QtWidgets.QAction('to global', menu_tr)
                qaction.cls_item = GlobalN
                qaction.triggered.connect(lambda: self.transfer_item(w_item))
                menu_tr.addAction(qaction)
                for global_cls in L_GLOBAL_CLS:
                    qaction = QtWidgets.QAction(f'to {global_cls.PREFIX:}',
                                                menu_tr)
                    qaction.cls_item = global_cls
                    qaction.triggered.connect(
                        lambda: self.transfer_item(w_item))
                    menu_tr.addAction(qaction)
            elif isinstance(obj, DataN):
                menu_tr = menu.addMenu("Transfer")
                qaction = QtWidgets.QAction('to data', menu_tr)
                qaction.cls_item = DataN
                qaction.triggered.connect(lambda: self.transfer_item(w_item))
                menu_tr.addAction(qaction)
                for data_cls in L_DATA_CLS:
                    qaction = QtWidgets.QAction(f'to {data_cls.PREFIX:}',
                                                menu_tr)
                    qaction.cls_item = data_cls
                    qaction.triggered.connect(
                        lambda: self.transfer_item(w_item))
                    menu_tr.addAction(qaction)

                qaction = QtWidgets.QAction('Duplicate', menu)
                qaction.triggered.connect(lambda: self.duplicate_data(w_item))
                menu.addAction(qaction)

            menu.exec_(self.viewport().mapToGlobal(position))

    def duplicate_data(self, widget: QtWidgets.QTreeWidgetItem):
        obj_data = widget.object
        widget_parent = widget.parent()
        obj_global = widget_parent.object
        text, ok = QtWidgets.QInputDialog.getText(
            self, f"Input dialog",
            f"Enter the new name of {obj_data.PREFIX:}")

        if ok:
            obj_data_new = obj_data.copy("".join(text.strip().split()))
            obj_global.add_items([obj_data_new])
            self.set_object_global()
        

    def rename_item(self, widget: QtWidgets.QTreeWidgetItem):
        """Rename item."""
        obj = widget.object
        text, ok = QtWidgets.QInputDialog.getText(
            self, f"Input dialog {obj.get_name():}", "Enter the new name")
        if ok:
            new_name = "".join(text.split())
            if isinstance(obj, GlobalN):
                obj.global_name = new_name
            elif isinstance(obj, DataN):
                obj.data_name = new_name
            elif isinstance(obj, LoopN):
                obj.loop_name = new_name
        self.set_object_global()

    def add_item(self, widget: QtWidgets.QTreeWidgetItem) \
            -> NoReturn:
        """Add object."""
        sender = self.sender()
        new_item = sender.cls_item()
        obj = widget.object
        if ((type(obj) is DataN) | (type(obj) is GlobalN)):
            item_cls = set([type(item) for item in obj.items])
            if type(new_item) not in item_cls:
                obj.CLASSES_OPTIONAL = tuple(list(obj.CLASSES_OPTIONAL) +
                                             [type(new_item)])
                obj.CLASSES = tuple(list(obj.CLASSES) + [type(new_item)])
        obj.add_items([new_item])
        self.set_object_global()

    def transfer_item(self, widget: QtWidgets.QTreeWidgetItem) \
            -> NoReturn:
        """Transfer to choosen item."""
        sender = self.sender()
        cls_item = sender.cls_item
        item_old = widget.object
        if cls_item is GlobalN:
            item_new = cls_item.make_container(
                (), tuple(set([type(item) for item in item_old])), "global")
        elif cls_item is DataN:
            item_new = cls_item.make_container(
                (), tuple(set([type(item) for item in item_old])), "data")
        else:
            item_new = cls_item()
        item_new.add_items(item_old.items)
        if isinstance(item_new, GlobalN):
            self.object = item_new
        else:
            widget_parent = widget.parent()
            item_parent = widget_parent.object
            item_parent.items.remove(item_old)
            item_parent.add_items([item_new])
        self.set_object_global()

    def add_items(self, widget: QtWidgets.QTreeWidgetItem) \
            -> NoReturn:
        """Add object."""
        obj = widget.object
        modal_widget = MWAddItem(self)
        modal_widget.set_object(obj)
        modal_widget.show()
        # obj.add_items([obj_new])

    def delete_object(self, widget: QtWidgets.QTreeWidgetItem) -> NoReturn:
        """Delete object."""
        widget_parent = widget.parent()
        obj = widget.object
        obj_parent = widget_parent.object
        if isinstance(obj_parent, (GlobalN, DataN)):
            obj_parent.items.remove(obj)
            widget_parent.removeChild(widget)

    def save_object(self, object_to_save) -> NoReturn:
        """Save as object."""
        f_name, okPressed = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Select a file:', "", "All files (*.*)")
        if not (okPressed):
            return None

        s_out = object_to_save.to_cif()

        with open(f_name, "w") as fid:
            fid.write(s_out)


def find_tree_item_position(
        w_main: QtWidgets.QTreeWidgetItem, child_item: QtWidgets.QTreeWidget)\
        -> int:
    """Find tree item position."""
    ind = w_main.indexOfTopLevelItem(child_item)
    if ind != -1:
        res = [ind]
    else:
        w_parent = child_item.parent()  # FIXME: it gives mistake
        ind_child = w_parent.indexOfChild(child_item)
        res = find_tree_item_position(w_main, w_parent)
        res.append(ind_child)
    return res


def make_tree_widget_item(widg: Union[QtWidgets.QTreeWidget,
                                      QtWidgets.QTreeWidgetItem],
                          object_item: Union[ItemN, LoopN, DataN, GlobalN],
                          name: str = None) \
        -> QtWidgets.QTreeWidgetItem:
    """Make tree widget item for item."""
    wi = QtWidgets.QTreeWidgetItem(widg)
    if name is None:
        wi.setText(0, f"{object_item.get_name():}")
    else:
        wi.setText(0, name)
    wi.object = object_item

    if object_item.is_defined():
        wi.setBackground(0, QtGui.QColor(255, 255, 255))
    else:
        wi.setBackground(0, QtGui.QColor(200, 110, 110))

    if isinstance(object_item, (GlobalN, DataN)):
        l_name = [item.get_name() for item in object_item.items]
        l_name_sort = sorted(l_name)
        for name in l_name_sort:
            ind = l_name.index(name)
            item = object_item.items[ind]
            wii = make_tree_widget_item(wi, item)
            wi.addChild(wii)
    elif isinstance(object_item, ItemN):
        for attr_name in object_item.ATTR_INT_NAMES:
            int_obj = getattr(object_item, attr_name)
            if isinstance(int_obj, (ItemN, LoopN)):
                wii = make_tree_widget_item(wi, int_obj, name=attr_name)
                wi.addChild(wii)
        for attr_name in object_item.ATTR_INT_PROTECTED_NAMES:
            if attr_name in object_item.__dict__.keys():
                int_obj = getattr(object_item, attr_name)
                if isinstance(int_obj, (ItemN, LoopN)):
                    wii = make_tree_widget_item(wi, int_obj, name=attr_name)
                    wi.addChild(wii)
    return wi


class MWAddItem(QtWidgets.QMainWindow):
    """MWAddItem class."""

    def __init__(self, parent=None):
        super(MWAddItem, self).__init__(parent)
        self.object = None
        self.setWindowTitle("Add Items")
        self.init_widgit()

    def init_widgit(self):
        """Init widgit."""
        widg = QtWidgets.QWidget(self)

        lay_central = QtWidgets.QVBoxLayout()

        self.w_te = QtWidgets.QTextEdit(widg)
        self.w_pb = QtWidgets.QPushButton(widg)
        self.w_pb.setText("Transform to items")
        self.w_pb.clicked.connect(self.add_items)

        lay_central.addWidget(self.w_te)
        lay_central.addWidget(self.w_pb)
        widg.setLayout(lay_central)
        self.setCentralWidget(widg)

    def set_object(self, obj):
        """Set object."""
        self.object = obj

    def add_items(self, obj):
        """Add items to object."""
        obj = self.object
        obj_cls = type(obj)
        s_cont = self.w_te.toPlainText()

        if isinstance(obj, GlobalN):
            l_item = (str_to_globaln(s_cont)).items
        elif isinstance(obj, DataN):
            l_item = str_to_items(s_cont)

        str_to_globaln

        cls_opt_obj = obj.CLASSES_OPTIONAL
        cls_obj = obj.CLASSES
        if ((obj_cls is DataN) | (obj_cls is GlobalN)):
            l_cls_item_add = [type(item) for item in l_item
                              if type(item) not in cls_obj]
            l_new_cls_opt_obj = list(cls_opt_obj) + l_cls_item_add
            l_new_cls_obj = list(cls_obj) + l_cls_item_add
            obj.CLASSES_OPTIONAL = tuple(l_new_cls_opt_obj)
            obj.CLASSES = tuple(l_new_cls_obj)
            obj.add_items(l_item)
        elif isinstance(obj, (DataN, GlobalN)):
            obj.add_items(l_item)
        self.close()



"""WFunction class."""
from PyQt5 import QtWidgets, QtCore, QtGui

from cryspy_editor.widgets.w_presentation import pass_func, run_function


class WBaseObj(QtWidgets.QGroupBox):
    """
    WBaseObj class.

    Attributes
    ----------
        - thread
    Mehtods
    -------
        - set_object
    """

    def __init__(self, parent=None):
        super(WBaseObj, self).__init__(parent)
        self.thread = None
        layout_central = QtWidgets.QVBoxLayout(self)
        # self.setWidgetResizable(True)
        self.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding))

        # self.wdoc = QtWidgets.QLabel(self)
        # self.wdoc.setSizePolicy(
        #     QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                           QtWidgets.QSizePolicy.Expanding))

        self.wlabel = QtWidgets.QTextEdit(self)
        self.wlabel.setAcceptRichText(True)
        self.wlabel.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding))
        self.wlabel.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
        # self.wlabel.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.wlabel.setAlignment(QtCore.Qt.AlignTop)
        # self.wlabel.setWordWrap(True)
        self.wlabel.setStyleSheet("background-color:white;")

        # layout_central.addWidget(self.wdoc,3)
        layout_central.addWidget(self.wlabel)

        pb = QtWidgets.QPushButton(self)
        pb.setText("Write to the object")
        pb.clicked.connect(self.write_to_object)
        layout_central.addWidget(pb)

        self.setLayout(layout_central)
        # self.setWidget(self.wlabel)

    def set_thread(self, thread: QtCore.QThread):
        """Set thread."""
        self.thread = thread

    def setText(self, text: str):
        """Set text."""
        self.wlabel.setText(text)

    def write_to_object(self):
        """Write to object."""
        if self.object is None:
            return
        stext = self.wlabel.toPlainText()
        obj2 = self.object.from_cif(stext)
        if obj2 is not None:
            self.object.copy_from(obj2)
        run_function(pass_func, (), self.thread)

    def set_object(self, obj):
        """Set text."""
        if obj is None:
            return
        self.object = obj
        self.wlabel.setText(obj.__repr__())
        self.wlabel.setToolTip(obj.__doc__)
        # self.wlabel.setHtml(obj._repr_html_())



"""Functions to create docks for dock-area."""
from typing import List
import numpy
from PyQt5 import QtWidgets, QtCore
import pyqtgraph as pg
import pyqtgraph.dockarea as pg_da

from cryspy import AtomSiteL, Crystal, Cell, SpaceGroup, \
    PdInstrResolution, Diffrn, \
    PdProcL, PdBackgroundL, Pd, PdMeasL, RhoChi, Chi2, RefineLs, \
    Pd2dMeas, Pd2dProc, Pd2d, value_error_to_string, \
    Setup, PhaseL, Range, DiffrnRadiation, Pd2dBackground, \
    Pd2dInstrResolution, DiffrnOrientMatrix, DiffrnReflnL, Phase, \
    AtomElectronConfigurationL, MEM, PdPeakL, Pd2dPeakL

ICON_SIZE = 50

pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')

pg_pen_black = pg.mkPen("k", width=5)
pg_pen_black_thick = pg.mkPen("k", width=2)


def w_for_presentation(obj, thread: QtCore.QThread) -> tuple:
    """Give tuple of docks for Dockarea."""
    if obj is None:
        return ()

    if isinstance(obj, RhoChi):
        docks = dock_rhochi(obj, thread)
        return docks
    elif isinstance(obj, MEM):
        docks = dock_mem(obj, thread)
        return docks
    elif isinstance(obj, Crystal):
        docks = dock_crystal(obj, thread)
        return docks
    elif isinstance(obj, Pd):
        docks = dock_pd(obj, thread)
        return docks
    elif isinstance(obj, Pd2d):
        docks = dock_pd2d(obj, thread)
        return docks
    elif isinstance(obj, Diffrn):
        docks = dock_diffrn(obj, thread)
        return docks
    return ()


def dock_rhochi(obj: RhoChi, thread: QtCore.QThread):
    """Dock for RhoChi object."""
    crystals = obj.crystals()
    experiments = obj.experiments()
    docks = []
    flag_crystals = len(crystals) != 0
    flag_experiments = len(experiments) != 0

    for experiment in experiments:
        if isinstance(experiment, Pd):
            docks.extend(list(dock_pd(experiment, thread)))
        elif isinstance(experiment, Pd2d):
            docks.extend(list(dock_pd2d(experiment, thread)))
        elif isinstance(experiment, Diffrn):
            docks.extend(list(dock_diffrn(experiment, thread)))

    if obj.is_defined():
        l_var_name = obj.get_variable_names()
        l_var_name_sigma = [tuple(list(name[:-1]) +
                                  [(f"{name[-1][0]:}_sigma", name[-1][1])])
                            for name in l_var_name]
        l_var_param = [obj.get_variable_by_name(name) for name in l_var_name]
        l_var_sigma = [obj.get_variable_by_name(name)
                       for name in l_var_name_sigma]
        l_var_str = [value_error_to_string(param, sigma)
                     for param, sigma in zip(l_var_param, l_var_sigma)]
        l_var_list = [f"{name[-1][0]:}: {param:}" for name, param in
                      zip(l_var_name, l_var_str)]
        if len(l_var_list) != 0:
            widget_list = QtWidgets.QListWidget()
            widget_list.addItems(l_var_list)
            docks.append(pg_da.Dock("Variables", widget=widget_list,
                                    autoOrientation=False))

    # Action doc
    layout_actions = QtWidgets.QHBoxLayout()
    if (flag_crystals & flag_experiments & obj.is_defined()):
        cb_1 = QtWidgets.QPushButton("Calc. Chi square")
        cb_1.clicked.connect(lambda: run_function(obj.calc_chi_sq,
                                                  (True, ), thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        # cb_1.setMaximumHeight(ICON_SIZE)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#AFD;")

        cb_2 = QtWidgets.QPushButton("Refine")
        cb_2.clicked.connect(lambda: run_function(obj.refine, (False, "BFGS",),
                                                  thread))
        cb_2.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        # cb_2.setMaximumHeight(ICON_SIZE)
        cb_2.setMinimumHeight(ICON_SIZE)
        cb_2.setStyleSheet("background-color:#DAF;")
        layout_actions.addWidget(cb_1)
        layout_actions.addWidget(cb_2)
    elif not(flag_crystals & flag_experiments):
        if not flag_crystals:
            cb_1 = QtWidgets.QPushButton("Add crytal")
            # cb_1.setStyleSheet("background-color:red;")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [
                Crystal(data_name="phase")], thread))
            layout_actions.addWidget(cb_1)

        if not flag_experiments:
            cb_2 = QtWidgets.QPushButton("Add diffrn")
            cb_2.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_2.setMaximumHeight(ICON_SIZE)
            cb_2.setMinimumHeight(ICON_SIZE)
            cb_2.clicked.connect(lambda: add_items(obj, [
                Diffrn(data_name="mono")], thread))
            cb_3 = QtWidgets.QPushButton("Add pd")
            cb_3.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_3.clicked.connect(lambda: add_items(obj, [
                Pd(data_name="powder1d")], thread))
            cb_3.setMaximumHeight(ICON_SIZE)
            cb_3.setMinimumHeight(ICON_SIZE)
            cb_4 = QtWidgets.QPushButton("Add pd2d")
            cb_4.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_4.setMaximumHeight(ICON_SIZE)
            cb_4.setMinimumHeight(ICON_SIZE)
            cb_4.clicked.connect(lambda: add_items(obj, [
                Pd2d(data_name="powder2d")], thread))
            layout_actions.addWidget(cb_2)
            layout_actions.addWidget(cb_3)
            layout_actions.addWidget(cb_4)
    else:
        qlabel = QtWidgets.QLabel(
            "To run calculations all items should be defined.")
        qlabel.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                             QtWidgets.QSizePolicy.Expanding)
        layout_actions.addWidget(qlabel)
    layout_actions.addStretch(1)

    widget_actions = QtWidgets.QWidget()
    widget_actions.setLayout(layout_actions)
    docks.append(pg_da.Dock("Actions RhoChi", widget=widget_actions,
                            autoOrientation=False))

    return tuple(docks)


def dock_mem(obj: MEM, thread: QtCore.QThread):
    """Dock for MEM object."""
    crystals = obj.crystals()
    experiments = obj.experiments()
    docks = []
    flag_crystals = len(crystals) != 0
    flag_experiments = len(experiments) != 0

    for experiment in experiments:
        docks.extend(list(dock_diffrn(experiment, thread)))

    # Action doc
    layout_actions = QtWidgets.QHBoxLayout()
    if (flag_crystals & flag_experiments & obj.is_defined()):
        if flag_crystals:
            crystal = crystals[0]
            if crystal.atom_electron_configuration is None:
                cb_1 = QtWidgets.QPushButton(
                    "Create AtomElectronConfiguration")
                cb_1.clicked.connect(lambda: crystal.add_items([
                    AtomElectronConfigurationL()]))
                cb_1.clicked.connect(lambda: run_function(pass_func, (),
                                                          thread))
                cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)
                cb_1.setMinimumHeight(ICON_SIZE)
                cb_1.setStyleSheet("background-color:#ADF;")
                layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Create prior density")
        cb_1.clicked.connect(lambda: run_function(obj.create_prior_density, (),
                                                  thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#DFA;")
        layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Calculate FR")
        cb_1.clicked.connect(lambda: run_function(obj.calc_fr, (),
                                                  thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#FAD;")
        layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Maximize entropy")
        cb_1.clicked.connect(lambda: run_function(obj.maximize_entropy, (),
                                                  thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#DAF;")
        layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Chi refinement")
        cb_1.clicked.connect(lambda: run_function(obj.refine_susceptibility,
                                                  (), thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#AFD;")
        layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Run cycle")
        cb_1.clicked.connect(lambda: run_function(obj.make_cycle, (),
                                                  thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#FDA;")
        layout_actions.addWidget(cb_1)

        cb_1 = QtWidgets.QPushButton("Save to '.den' files")
        cb_1.clicked.connect(lambda: run_function(obj.save_to_file_den, (),
                                                  thread))
        cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding)
        cb_1.setMinimumHeight(ICON_SIZE)
        cb_1.setStyleSheet("background-color:#DFD;")
        layout_actions.addWidget(cb_1)
    elif not(flag_crystals & flag_experiments):
        if not flag_crystals:
            cb_1 = QtWidgets.QPushButton("Add crytal")
            # cb_1.setStyleSheet("background-color:red;")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [
                Crystal(data_name="phase")], thread))
            layout_actions.addWidget(cb_1)

        if not flag_experiments:
            cb_2 = QtWidgets.QPushButton("Add diffrn")
            cb_2.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_2.setMaximumHeight(ICON_SIZE)
            cb_2.setMinimumHeight(ICON_SIZE)
            cb_2.clicked.connect(lambda: add_items(obj, [
                Diffrn(data_name="mono")], thread))
            layout_actions.addWidget(cb_2)
    else:
        qlabel = QtWidgets.QLabel(
            "To run calculations all items should be defined.")
        qlabel.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                             QtWidgets.QSizePolicy.Expanding)
        layout_actions.addWidget(qlabel)
    layout_actions.addStretch(1)

    widget_actions = QtWidgets.QWidget()
    widget_actions.setLayout(layout_actions)
    docks.append(pg_da.Dock("Actions MEM", widget=widget_actions,
                            autoOrientation=False))

    return tuple(docks)


def pass_func():
    return


def add_items(obj, l_item, thread: QtCore.QThread):
    obj.add_items(l_item)
    thread.function = pass_func
    thread.arguments = ()
    thread.start()

def run_function(func, args, thread: QtCore.QThread):
    """Run function."""
    thread.function = func
    thread.arguments = args
    thread.start()


def dock_pd(obj: Pd, thread: QtCore.QThread):  # FIXME: dock_peak should be added
    """
    Form dock_pd.

    Based on
    --------
        - dock_proc
        - dock_meas
        - dock_chi2
        - dock_refine_ls
        - dock_peak
    """
    f_proc, f_meas, f_peak = False, False, False
    proc, meas = None, None
    meas = obj.pd_meas
    f_meas = meas is not None
    chi2 = obj.chi2
    f_chi2 = chi2 is not None
    refine_ls = obj.refine_ls
    f_refine_ls = refine_ls is not None

    proc = obj.pd_proc
    f_proc = proc is not None
    phase = obj.phase
    f_phase = phase is not None
    l_pd_peak = []
    if f_phase:
        for item in phase.items:
            pd_peak = getattr(obj, f"pd_peak_{item.label.lower():}")
            if pd_peak is not None:
                l_pd_peak.append(pd_peak)

    docks = []

    f_setup = obj.setup is not None
    f_pd_instr_resolution = obj.pd_instr_resolution is not None
    f_pd_background = obj.pd_background is not None
    f_range = obj.range is not None
    # f_diffrn_radiation = obj.diffrn_radiation is not None
    
    if not(f_chi2 & f_meas & f_setup & f_pd_instr_resolution & f_phase &
           f_pd_background & f_range):
        layout_actions = QtWidgets.QHBoxLayout()
        if not f_chi2:
            cb_1 = QtWidgets.QPushButton("Add chi2")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Chi2()], thread))
            layout_actions.addWidget(cb_1)
        if not f_meas:
            cb_1 = QtWidgets.QPushButton("Add pd_meas")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [PdMeasL()], thread))
            layout_actions.addWidget(cb_1)
        if not f_setup:
            cb_1 = QtWidgets.QPushButton("Add setup")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            layout_actions.addWidget(cb_1)
        if not f_pd_instr_resolution:
            cb_1 = QtWidgets.QPushButton("Add pd_instr_resolution")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [PdInstrResolution()],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if not f_phase:
            cb_1 = QtWidgets.QPushButton("Add phase")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            vv = PhaseL()
            vv.items = [Phase(label="phase", igsize=0., scale=1.)]
            cb_1.clicked.connect(lambda: add_items(obj, [vv], thread))
            layout_actions.addWidget(cb_1)
        if not f_pd_background:
            cb_1 = QtWidgets.QPushButton("Add pd_background")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [PdBackgroundL()],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if not f_range:
            cb_1 = QtWidgets.QPushButton("Add range")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(
                obj, [Range(ttheta_min=2, ttheta_max=100.)], thread))
            layout_actions.addWidget(cb_1)
        # if not f_diffrn_radiation:
        #     cb_1 = QtWidgets.QPushButton("Add diffrn_radiation")
        #     cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                        QtWidgets.QSizePolicy.Expanding)
        #     cb_1.clicked.connect(lambda: add_items(obj, [DiffrnRadiation()], thread))
        #     layout_actions.addWidget(cb_1)
        layout_actions.addStretch(1)
        widget_actions = QtWidgets.QWidget()
        widget_actions.setLayout(layout_actions)
        docks.append(pg_da.Dock("Actions pd", widget=widget_actions,
                                autoOrientation=False))

    if f_proc:
        docks = docks + list(dock_proc(proc, l_pd_peak))
    elif f_meas:
        docks = docks + list(dock_meas(meas))

    if f_chi2:
        docks = docks + list(dock_chi2(chi2))
    if f_refine_ls:
        docks = docks + list(dock_refine_ls(refine_ls))

    return tuple(docks)


def dock_proc(obj: PdProcL, l_pd_peak: List[PdPeakL]):
    """Form dock for PdProc object."""
    np_x_1 = numpy.array(obj.ttheta, dtype=float)
    np_y_u_1 = numpy.array(obj.intensity_up, dtype=float)
    # np_y_su_1 = numpy.array(obj.intensity_up_sigma, dtype=float)
    np_y_d_1 = numpy.array(obj.intensity_down, dtype=float)
    # np_y_sd_1 = numpy.array(obj.intensity_down_sigma, dtype=float)
    # np_y_b_1 = numpy.array(obj.intensity_bkg_calc, dtype=float)
    # np_y_u_2 = numpy.array(obj.intensity_up_total, dtype=float)
    # np_y_d_2 = numpy.array(obj.intensity_down_total, dtype=float)
    np_y_s_1 = numpy.array(obj.intensity, dtype=float)
    np_y_ss_1 = numpy.array(obj.intensity_sigma, dtype=float)
    np_y_m_1 = np_y_u_1 - np_y_d_1
    np_y_sm_1 = np_y_ss_1
    np_y_s_2 = numpy.array(obj.intensity_total, dtype=float)
    np_y_m_2 = numpy.array(obj.intensity_diff_total, dtype=float)

    # np_x_2 = numpy.array(peak.ttheta, dtype = float)
    # np_xysm_1 = numpy.vstack((np_x_1, np_y_s_1, np_y_ss_1,
    #                           np_y_s_2)).transpose()
    # np_xb_1 = numpy.vstack((np_x_1, 2*np_y_b_1)).transpose()

    vb_1 = pg.ViewBox()
    vb_1.setMouseMode(vb_1.RectMode)
    widget = pg.PlotWidget(viewBox=vb_1)
    p1 = widget.plotItem
    p2 = add_2nd_axis_to_plot(p1, label_axis_1="Intensity (a.u.)",
                              label_axis_2="Difference")
    err1 = pg.ErrorBarItem(x=np_x_1, y=np_y_s_1, top=np_y_ss_1,
                           bottom=np_y_ss_1, beam=0.5)
    p1.plot(x=np_x_1, y=np_y_s_1, name="exp", title="model",
            symbol='o', pen={'color': 0.8, 'width': 2})
    p1.plot(x=np_x_1, y=np_y_s_2, name="model", title="model",
            pen=pg_pen_black)
    p1.addItem(err1)
    p2.addItem(pg.PlotCurveItem(x=np_x_1, y=np_y_s_1-np_y_s_2,
                                pen=pg_pen_black_thick))

    vb_2 = pg.ViewBox()
    vb_2.setMouseMode(vb_2.RectMode)
    widget_2 = pg.PlotWidget(viewBox=vb_2)
    p3 = widget_2.plotItem
    p4 = add_2nd_axis_to_plot(p3, label_axis_1="Intensity (a.u.)",
                              label_axis_2="Difference")
    err1 = pg.ErrorBarItem(x=np_x_1, y=np_y_m_1, top=np_y_sm_1,
                           bottom=np_y_ss_1, beam=0.5)
    p3.plot(x=np_x_1, y=np_y_m_1, name="exp", title="model",
            symbol='o', pen={'color': 0.8, 'width': 2})
    p3.plot(x=np_x_1, y=np_y_m_2, name="model", title="model",
            pen=pg_pen_black)
    p3.addItem(err1)
    p4.addItem(pg.PlotCurveItem(x=np_x_1, y=np_y_m_1-np_y_m_2,
                                pen=pg_pen_black_thick))

    [x_range_12, y_range_12] = p1.viewRange()
    y_range = y_range_12[1]-y_range_12[0]
    p2.setRange(yRange=(-0.15*y_range, 0.85*y_range))
    [x_range_34, y_range_34] = p3.viewRange()
    y_range = y_range_34[1]-y_range_34[0]
    p3.setYRange(-0.50*y_range, 0.50*y_range)
    p4.setRange(yRange=(-0.15*y_range, 0.85*y_range))

    yrange_2 = p2.viewRange()[1]
    ysize_2 = 0.05*(yrange_2[1]-yrange_2[0])
    yshift_2 = 0.5*0.5*0.5*ysize_2*len(l_pd_peak)
    yrange_4 = p4.viewRange()[1]
    ysize_4 = 0.05*(yrange_4[1]-yrange_4[0])
    yshift_4 = 0.5*ysize_4*len(l_pd_peak)
    for pd_peak in l_pd_peak:
        numpy_ttheta = numpy.array(pd_peak.ttheta, dtype=float)
        np_zero = 0.*numpy_ttheta
        ind_miller_2 = pg.ErrorBarItem(
            x=numpy_ttheta, y=np_zero+yshift_2, top=np_zero+yshift_2+ysize_2,
            bottom=np_zero-yshift_2+ysize_2, beam=0.0)
        ind_miller_4 = pg.ErrorBarItem(
            x=numpy_ttheta, y=np_zero+yshift_4, top=np_zero+yshift_4+ysize_4,
            bottom=np_zero-yshift_4+ysize_4, beam=0.0)
        p2.addItem(ind_miller_2)
        p4.addItem(ind_miller_4)
        yshift_2 -= 0.5*ysize_2
        yshift_4 -= 1.0*ysize_4


    dock_1 = pg_da.Dock("Sum", widget=widget, autoOrientation=False)
    dock_2 = pg_da.Dock("Difference", widget=widget_2, autoOrientation=False)

    return (dock_1, dock_2, )


def dock_meas(obj: PdMeasL):
    """Form docks for PdMeasL object."""
    np_x_1 = numpy.array(obj.ttheta, dtype=float)
    np_y_u_1 = numpy.array(obj.intensity_up, dtype=float)
    np_y_su_1 = numpy.array(obj.intensity_up_sigma, dtype=float)
    np_y_d_1 = numpy.array(obj.intensity_down, dtype=float)
    np_y_sd_1 = numpy.array(obj.intensity_down_sigma, dtype=float)

    if numpy.all(numpy.isnan(np_y_u_1)):
        np_y_s_1 = numpy.array(obj.intensity, dtype=float)
        np_y_ss_1 = numpy.array(obj.intensity_sigma, dtype=float)
        np_y_m_1 = 0*np_y_s_1
        np_y_sm_1 = np_y_ss_1
    else:
        np_y_s_1 = np_y_u_1 + np_y_d_1
        np_y_ss_1 = numpy.sqrt(numpy.square(np_y_su_1) +
                               numpy.square(np_y_sd_1))
        np_y_m_1 = np_y_u_1 - np_y_d_1
        np_y_sm_1 = np_y_ss_1

    vb_1 = pg.ViewBox()
    vb_1.setMouseMode(vb_1.RectMode)
    widget = pg.PlotWidget(viewBox=vb_1)
    p1 = widget.plotItem
    p2 = add_2nd_axis_to_plot(p1, label_axis_1="Intensity (a.u.)",
                              label_axis_2="Intensity (a.u.)")
    err_s = pg.ErrorBarItem(x=np_x_1, y=np_y_s_1, top=np_y_ss_1,
                            bottom=np_y_ss_1, beam=0.5)
    err_m = pg.ErrorBarItem(x=np_x_1, y=np_y_m_1, top=np_y_sm_1,
                            bottom=np_y_sm_1, beam=0.5)
    p1.plot(x=np_x_1, y=np_y_s_1, name="sum", title="Sum",
            symbol='o', pen={'color': 0.8, 'width': 2})
    p1.addItem(err_s)
    p2.addItem(pg.PlotCurveItem(x=np_x_1, y=np_y_m_1, name="diff",
                                title="Difference", pen=pg_pen_black_thick))
    p2.addItem(pg.ScatterPlotItem(x=np_x_1, y=np_y_m_1,
                                  symbol='o', pen={'color': 0.8, 'width': 2}))
    p2.addItem(err_m)
    [x_range_12, y_range_12] = p1.viewRange()
    y_range = y_range_12[1]-y_range_12[0]
    p2.setRange(yRange=(-0.25*y_range, 0.75*y_range))
    
    vb_2 = pg.ViewBox()
    vb_2.setMouseMode(vb_2.RectMode)
    widget_2 = pg.PlotWidget(viewBox=vb_2)
    p3 = widget_2.plotItem
    p4 = add_2nd_axis_to_plot(p3, label_axis_1="Intensity (a.u.)",
                              label_axis_2="Intensity (a.u.)")
    err_u = pg.ErrorBarItem(x=np_x_1, y=np_y_u_1, top=np_y_su_1,
                            bottom=np_y_su_1, beam=0.5)
    err_d = pg.ErrorBarItem(x=np_x_1, y=np_y_d_1, top=np_y_sd_1,
                            bottom=np_y_sd_1, beam=0.5)
    p3.plot(x=np_x_1, y=np_y_u_1, name="up", title="Up",
            symbol='o', pen={'color': 0.8, 'width': 2})
    p3.addItem(err_u)
    p4.addItem(pg.PlotCurveItem(x=np_x_1, y=np_y_d_1, name="down",
                                title="Down", pen=pg_pen_black_thick))
    p4.addItem(pg.ScatterPlotItem(x=np_x_1, y=np_y_d_1,
                                  symbol='o', pen={'color': 0.8, 'width': 2}))
    p4.addItem(err_d)

    dock_1 = pg_da.Dock("Sum/Diff.", widget=widget, autoOrientation=False)
    dock_2 = pg_da.Dock("Up/Down", widget=widget_2, autoOrientation=False)
    return (dock_1, dock_2, )


def dock_chi2(obj: Chi2):
    """Form docks for Chi2 object."""
    docks = []
    layout = QtWidgets.QGridLayout()
    cb_s = QtWidgets.QCheckBox("sum")
    if obj.sum is not None:
        cb_s.setCheckState(2*int(obj.sum))
    cb_s.stateChanged.connect(lambda x: setattr(obj, "sum", bool(x/2)))
    layout.addWidget(cb_s, 0, 0)

    cb_m = QtWidgets.QCheckBox("diff")
    if obj.diff is not None:
        cb_m.setCheckState(2*int(obj.diff))
    cb_m.stateChanged.connect(lambda x: setattr(obj, "diff", bool(x/2)))
    layout.addWidget(cb_m, 1, 0)

    cb_u = QtWidgets.QCheckBox("up")
    if obj.up is not None:
        cb_u.setCheckState(2*int(obj.up))
    cb_u.stateChanged.connect(lambda x: setattr(obj, "up", bool(x/2)))
    layout.addWidget(cb_u, 0, 1)

    cb_d = QtWidgets.QCheckBox("down")
    if obj.down is not None:
        cb_d.setCheckState(2*int(obj.down))
    cb_d.stateChanged.connect(lambda x: setattr(obj, "down", bool(x/2)))
    layout.addWidget(cb_d, 1, 1)

    widget = QtWidgets.QWidget()
    widget.setLayout(layout)
    dock_1 = pg_da.Dock("Chi2", widget=widget, autoOrientation=False)
    docks.append(dock_1)
    return tuple(docks)


def dock_refine_ls(obj: RefineLs):
    """Form docks for RefineLs object."""
    docks = []
    if ((obj.number_reflns is not None) &
            (obj.goodness_of_fit_all is not None)):
        val_1, val_2 = obj.number_reflns, obj.goodness_of_fit_all
        text = f"Points: {val_1:}\nGoF: {val_2:.2f}"
        widget = QtWidgets.QLabel(text)
        docks.append(pg_da.Dock("RefineLs", widget=widget,
                                autoOrientation=False))
    return tuple(docks)


def matrix_widget(win:pg.GraphicsLayoutWidget,
                  data, x_step, y_step, x_point_0, y_point_0):
    win.clear()
    # win.setWindowTitle('pyqtgraph example: Image Analysis')

    # A plot area (ViewBox + axes) for displaying the image
    p1 = win.addPlot(title="")

    # Item for displaying image data
    img = pg.ImageItem()
    p1.addItem(img)

    # # Custom ROI for selecting an image region
    # roi = pg.ROI([-8, 14], [6, 5])
    # roi.addScaleHandle([0.5, 1], [0.5, 0.5])
    # roi.addScaleHandle([0, 0.5], [0.5, 0.5])
    # p1.addItem(roi)
    # roi.setZValue(10)  # make sure ROI is drawn above image

    # Isocurve drawing
    iso = pg.IsocurveItem(level=0.8, pen=pg.mkPen("g", width=2))
    iso.setParentItem(img)
    iso.setZValue(5)

    # Contrast/color control
    hist = pg.HistogramLUTItem()
    hist.setImageItem(img)
    hist.setHistogramRange(data.mean()-3*numpy.std(data), data.mean()+3*numpy.std(data))
    win.addItem(hist)

    # Draggable line for setting isocurve level
    isoLine = pg.InfiniteLine(angle=0, movable=True, pen=pg.mkPen("g", width=2))
    hist.vb.addItem(isoLine)
    hist.vb.setMouseEnabled(y=True) # makes user interaction a little easier
    isoLine.setValue(0.8)
    isoLine.setZValue(1000) # bring iso line above contrast controls

    # # Another plot area for displaying ROI data
    # win.nextRow()
    # p2 = win.addPlot(colspan=2)
    # p2.setMaximumHeight(250)
    # win.resize(800, 800)
    # win.show()

    # Generate image data
    img.setImage(data)
    hist.setLevels(data.mean()-1*numpy.std(data),
                   data.mean()+1*numpy.std(data))

    # build isocurves from smoothed data
    iso.setData(pg.gaussianFilter(data, (2, 2)))

    # set position and scale of image
    img.scale(x_step, y_step)
    img.translate(x_point_0, y_point_0)

    # zoom to fit imageo
    p1.autoRange()

    # # Callbacks for handling user interaction
    # def updatePlot():
    #     selected = roi.getArrayRegion(data, img)
    #     p2.plot(selected.mean(axis=0), clear=True)

    # roi.sigRegionChanged.connect(updatePlot)
    # updatePlot()

    def updateIsocurve():
        iso.setLevel(isoLine.value())

    isoLine.sigDragged.connect(updateIsocurve)

    def imageHoverEvent(event):
        """Show the position, pixel, and value under the mouse cursor."""
        if event.isExit():
            p1.setTitle("")
            return
        pos = event.pos()
        i, j = pos.y(), pos.x()
        i = int(numpy.clip(i, 0, data.shape[0] - 1))
        j = int(numpy.clip(j, 0, data.shape[1] - 1))
        val = data[i, j]
        ppos = img.mapToParent(pos)
        x, y = ppos.x(), ppos.y()
        p1.setTitle("pos: (%0.1f, %0.1f)  pixel: (%d, %d)  value: %g" % (
            x, y, i, j, val))

    # Monkey-patch the image to use our custom hover function.
    # This is generally discouraged (you should subclass ImageItem instead),
    # but it works for a very simple use like this.
    img.hoverEvent = imageHoverEvent
    return

def add_2nd_axis_to_plot(p1:pg.PlotItem, label_axis_1:str="",
                         label_axis_2:str=""):
    p1.setLabels(left=label_axis_1)
    
    ## create a new ViewBox, link the right axis to its coordinate system
    p2 = pg.ViewBox()
    # p2.setMouseMode(p2.RectMode)
    p1.showAxis('right')
    p1.scene().addItem(p2)
    p1.getAxis('right').linkToView(p2)
    p2.setXLink(p1)
    p1.getAxis('right').setLabel(label_axis_2, color='#0000ff')

    ## Handle view resizing
    def updateViews():
        ## view has resized; update auxiliary views to match
        p2.setGeometry(p1.vb.sceneBoundingRect())
        ## need to re-update linked axes since this was called
        ## incorrectly while views had different shapes.
        ## (probably this should be handled in ViewBox.resizeEvent)
        p2.linkedViewChanged(p1.vb, p2.XAxis)
    updateViews()
    p1.vb.sigResized.connect(updateViews)
    return p2

def dock_pd_instr_resolution(obj:PdInstrResolution):
    tth = numpy.linspace(1, 120, 119)
    docks = []
    if obj.is_defined:
        h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = obj.calc_resolution(tth)

        widget = pg.GraphicsLayoutWidget(title="sum/diff")
        p1 = widget.addPlot(x=tth, y=h_g, name="h_g", title="Resolution",
                pen=pg_pen_black)
        p1.plot(x=tth, y=h_l, name="h_l", pen=pg_pen_black)
        docks.append(pg_da.Dock("Instrumental resolution", widget=widget,
                                autoOrientation=False))
    return docks


def dock_cell(obj:Cell):
    docks = []
    s_str = obj.report()
    if s_str is not None:
        widget = QtWidgets.QLabel()
        area = QtWidgets.QScrollArea()
        area.setWidgetResizable(True)
        area.setWidget(widget)

        widget.setStyleSheet("background-color:white;")
        #widget.setFont(QtGui.QFont("Courier", 12, QtGui.QFont.Bold))
        widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                             QtWidgets.QSizePolicy.Expanding)
        widget.setText(s_str)

        dock_1 = pg_da.Dock("Cell", widget=area, autoOrientation=False)
        docks.append(dock_1)
    return tuple(docks)

def dock_space_group(obj:SpaceGroup):
    docks = []
    s_str = obj.report_space_group()
    if s_str is not None:
        widget = QtWidgets.QLabel()
        area = QtWidgets.QScrollArea()
        area.setWidgetResizable(True)
        area.setWidget(widget)

        widget.setStyleSheet("background-color:white;")
        #widget.setFont(QtGui.QFont("Courier", 12, QtGui.QFont.Bold))
        widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        widget.setText(s_str)

        dock_1 = pg_da.Dock("Space Group", widget=area, autoOrientation=False)
        docks.append(dock_1)
    return tuple(docks)

def dock_crystal(obj:Crystal, thread: QtCore.QThread):
    docks = []
    cell = obj.cell
    space_group = obj.space_group
    if not obj.is_defined():
        layout_actions = QtWidgets.QHBoxLayout()
        if cell is None:
            cb_1 = QtWidgets.QPushButton("Add cell")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, 
                                                   [Cell()], thread))
            layout_actions.addWidget(cb_1)
        if space_group is None:
            cb_1 = QtWidgets.QPushButton("Add space group")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, 
                                                   [SpaceGroup(it_number=1)],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if obj.atom_site is None:
            cb_1 = QtWidgets.QPushButton("Add atom site")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, 
                                                    [AtomSiteL()], thread))
            layout_actions.addWidget(cb_1)
        layout_actions.addStretch(1)
        widget_actions = QtWidgets.QWidget()
        widget_actions.setLayout(layout_actions)
        docks.append(pg_da.Dock("Actions crystal", widget=widget_actions,
                                autoOrientation=False))

    if cell is not None:
        docks.extend(list(dock_cell(cell)))
    if space_group is not None:
        docks.extend(list(dock_space_group(space_group)))
    s_str = obj.report_main_axes_of_magnetization_ellipsoids()
    if s_str is not None:
        widget = QtWidgets.QLabel()
        area = QtWidgets.QScrollArea()
        area.setWidgetResizable(True)
        area.setWidget(widget)

        widget.setStyleSheet("background-color:white;")
        #widget.setFont(QtGui.QFont("Courier", 12, QtGui.QFont.Bold))
        widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        widget.setText(s_str)

        dock_1 = pg_da.Dock("Magnetization", widget=area,
                            autoOrientation=False)
        docks.append(dock_1)
    return tuple(docks)


def dock_pd2d(obj:Pd2d, thread: QtCore.QThread):
    f_proc, f_meas, f_bkgr, f_peak = False, False, False, False
    proc, meas, bkgr, peak = None, None, None, None
    meas = obj.pd2d_meas
    f_meas = meas is not None
    bkgr = obj.pd2d_background
    f_bkgr = bkgr is not None
    chi2 = obj.chi2
    f_chi2 = chi2 is not None
    refine_ls = obj.refine_ls
    f_refine_ls = refine_ls is not None

    proc = obj.pd2d_proc
    f_proc = proc is not None
    l_pd2d_peak = []
    phase = obj.phase
    for item in phase.items:
        pd2d_peak = getattr(obj, f"pd2d_peak_{item.label.lower():}")
        if pd2d_peak is not None:
            l_pd2d_peak.append(pd2d_peak)

    docks = []

    f_setup = obj.setup is not None
    f_pd2d_instr_resolution = obj.pd2d_instr_resolution is not None
    f_phase = obj.phase is not None
    f_pd2d_background = obj.pd2d_background is not None
    f_range = obj.range is not None
    # f_diffrn_radiation = obj.diffrn_radiation is not None

    if not(f_chi2 & f_meas & f_setup & f_pd2d_instr_resolution & f_phase &
           f_pd2d_background & f_range):
        layout_actions = QtWidgets.QHBoxLayout()
        if not f_chi2:
            cb_1 = QtWidgets.QPushButton("Add chi2")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Chi2()], thread))
            layout_actions.addWidget(cb_1)
        if not f_meas:
            cb_1 = QtWidgets.QPushButton("Add pd2d_meas")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Pd2dMeas()], thread))
            layout_actions.addWidget(cb_1)
        if not f_setup:
            cb_1 = QtWidgets.QPushButton("Add setup")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            layout_actions.addWidget(cb_1)
        if not f_pd2d_instr_resolution:
            cb_1 = QtWidgets.QPushButton("Add \npd2d_instr_resolution")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(
                obj, [Pd2dInstrResolution()], thread))
            layout_actions.addWidget(cb_1)
        if not f_phase:
            cb_1 = QtWidgets.QPushButton("Add phase")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            vv = PhaseL()
            vv.items = [Phase(label="phase", igsize=0., scale=1.)]
            cb_1.clicked.connect(lambda: add_items(obj, [vv], thread))
            layout_actions.addWidget(cb_1)
        if not f_pd2d_background:
            cb_1 = QtWidgets.QPushButton("Add \npd2d_background")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Pd2dBackground()],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if not f_range:
            cb_1 = QtWidgets.QPushButton("Add range")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(
                obj, [Range(ttheta_min=4., ttheta_max=80., phi_min=-10.,
                            phi_max=20.)], thread))
            layout_actions.addWidget(cb_1)
        layout_actions.addStretch(1)
        # if not f_diffrn_radiation:
        #     cb_1 = QtWidgets.QPushButton("Add diffrn_radiation")
        #     cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        #                        QtWidgets.QSizePolicy.Expanding)
        #     cb_1.clicked.connect(lambda: add_items(obj, [DiffrnRadiation()],
        #                                            thread))
        #     layout_actions.addWidget(cb_1)
        widget_actions = QtWidgets.QWidget()
        widget_actions.setLayout(layout_actions)
        docks.append(pg_da.Dock("Actions pd2d", widget=widget_actions,
                                autoOrientation=False))

    if f_proc:
        docks.extend(list(dock_pd2d_proc(proc, l_pd2d_peak)))
    elif f_meas:
        docks.extend(list(dock_pd2d_meas(meas)))
    if f_chi2:
        docks.extend(list(dock_chi2(chi2)))
    if f_refine_ls:
        docks.extend(list(dock_refine_ls(refine_ls)))
    return tuple(docks)


def dock_pd2d_meas(obj:Pd2dMeas):
    docks = []
    ttheta = obj.ttheta
    phi = obj.phi
    intensity_up = obj.intensity_up
    intensity_up_sigma = obj.intensity_up_sigma
    intensity_down = obj.intensity_down
    intensity_down_sigma = obj.intensity_down_sigma
    if ((intensity_up is None) & (intensity_down is None)):
        return tuple(docks)
    int_s =  numpy.nan_to_num(intensity_up + intensity_down)
    int_m =  numpy.nan_to_num(intensity_up - intensity_down)
    ttheta_step = ttheta[1]-ttheta[0]
    phi_step = phi[1]-phi[0]
    data = int_m

    x_step = ttheta_step
    y_step = phi_step
    x_point_0 = int(ttheta[0]/ttheta_step)
    y_point_0 = int(phi[0]/phi_step)

    win = pg.GraphicsLayoutWidget()

    win_1d = pg.GraphicsLayoutWidget()

    z_sum_e = intensity_up.transpose()+intensity_down.transpose()
    # z_sum_s_sq = (intensity_up_sigma.transpose())**2+(intensity_down_sigma.transpose())**2
    _z_1_e = numpy.where(numpy.isnan(z_sum_e), 0., z_sum_e).sum(axis=0)
    # _z_1_s = numpy.sqrt(numpy.where(numpy.isnan(z_sum_e), 0., z_sum_s_sq).sum(axis=0))
    _n_1 = numpy.where(numpy.isnan(z_sum_e), 0., 1.).sum(axis=0)
    z_1_sum_e = _z_1_e/_n_1
    # z_1_sum_s = _z_1_s/_n_1

    z_diff_e = intensity_up.transpose()-intensity_down.transpose()
    _z_1_e = numpy.where(numpy.isnan(z_diff_e), 0., z_diff_e).sum(axis=0)
    _n_1 = numpy.where(numpy.isnan(z_diff_e), 0., 1.).sum(axis=0)
    z_1_diff_e = _z_1_e/_n_1
    # z_1_diff_s = _z_1_s/_n_1

    p1 = win_1d.addPlot()

    vb_p2 = add_2nd_axis_to_plot(p1, label_axis_1="Intensity (a.u.)",
                                  label_axis_2="Intensity (a.u.)")

    p1.plot(ttheta, z_1_sum_e, symbol='o', pen={'color': 0.8, 'width': 2})
    p2 = pg.ScatterPlotItem(ttheta, z_1_diff_e, symbol='o',
                                pen={'color': 0.2, 'width': 2})
    p3 = pg.PlotCurveItem(x=ttheta, y=z_1_diff_e, pen=pg_pen_black_thick)
    vb_p2.addItem(p2)
    vb_p2.addItem(p3)

    [x_range_12, y_range_12] = p1.viewRange()
    y_range = y_range_12[1]-y_range_12[0]
    vb_p2.setRange(yRange=(-0.25*y_range, 0.75*y_range))

    cb_s = QtWidgets.QPushButton("2D sum")
    cb_s.clicked.connect(lambda: matrix_widget(win, int_s, x_step, y_step,
                                               x_point_0, y_point_0))

    cb_m = QtWidgets.QPushButton("2D diff")
    cb_m.clicked.connect(lambda: matrix_widget(win, int_m, x_step, y_step,
                                               x_point_0, y_point_0))

    matrix_widget(win, int_m, x_step, y_step, x_point_0, y_point_0)

    layout = QtWidgets.QGridLayout()
    layout.addWidget(cb_s, 0, 0)
    layout.addWidget(cb_m, 0, 1)
    layout.addWidget(win, 1, 0, 1, 2)
    widget = QtWidgets.QWidget()
    widget.setLayout(layout)
    #win = matrix_widget(data, x_step, y_step, x_point_0, y_point_0)

    #docks.append(pg_da.Dock("2D measurements: sum", widget=imv_s))

    docks.append(pg_da.Dock("Projection of sum and difference", widget=win_1d,
                            autoOrientation=False))
    docks.append(pg_da.Dock("Plot", widget=widget, autoOrientation=False))
    return tuple(docks)


def dock_pd2d_proc(obj:Pd2dProc, l_pd2d_peak: List[Pd2dPeakL]):
    """Docks for pd2d_proc object."""
    docks = []
    ttheta = obj.ttheta
    phi = obj.phi

    intensity_up = obj.intensity_up
    intensity_up_sigma = obj.intensity_up_sigma
    intensity_down = obj.intensity_down
    intensity_down_sigma = obj.intensity_down_sigma
    intensity_up_total = obj.intensity_up_total
    intensity_down_total = obj.intensity_down_total

    if ((intensity_up_total is None) & (intensity_down_total is None)):
        return tuple(docks)


    z_sum = intensity_up_total+intensity_down_total
    z_sum_e = intensity_up+intensity_down
    z_sum_s_sq = (intensity_up_sigma)**2+(intensity_down_sigma)**2
    _z_1_e = numpy.where(numpy.isnan(z_sum_e), 0., z_sum_e).sum(axis=1)
    _z_1 = numpy.where(numpy.isnan(z_sum_e), 0., z_sum).sum(axis=1)
    _z_1_s = numpy.sqrt(numpy.where(numpy.isnan(z_sum_e), 0., z_sum_s_sq).sum(axis=1))
    _n_1 = numpy.where(numpy.isnan(z_sum_e), 0., 1.).sum(axis=1)
    z_1_sum = _z_1/_n_1
    z_1_sum_e = _z_1_e/_n_1
    z_1_sum_s = _z_1_s/_n_1


    z_diff = intensity_up_total-intensity_down_total
    z_diff_e = intensity_up-intensity_down
    _z_1_e = numpy.where(numpy.isnan(z_diff_e), 0., z_diff_e).sum(axis=1)
    _z_1 = numpy.where(numpy.isnan(z_diff_e), 0., z_diff).sum(axis=1)
    _n_1 = numpy.where(numpy.isnan(z_diff_e), 0., 1.).sum(axis=1)
    z_1_diff = _z_1/_n_1
    z_1_diff_e = _z_1_e/_n_1
    z_1_diff_s = _z_1_s/_n_1

    z_sum_e_0 = numpy.nan_to_num(z_sum_e)
    z_diff_e_0 = numpy.nan_to_num(z_diff_e)

    ttheta_step = ttheta[1]-ttheta[0]
    phi_step = phi[1]-phi[0]
    x_step = ttheta_step
    y_step = phi_step
    x_point_0 = int(ttheta[0]/ttheta_step)
    y_point_0 = int(phi[0]/phi_step)

    win = pg.GraphicsLayoutWidget()

    err_s = pg.ErrorBarItem(x=ttheta, y=z_1_sum_e, top=z_1_sum_s,
                                      bottom=z_1_sum_s, beam=0.5)
    err_m = pg.ErrorBarItem(x=ttheta, y=z_1_diff_e, top=z_1_diff_s,
                                      bottom=z_1_diff_s, beam=0.5)


    win_1d_s = pg.GraphicsLayoutWidget()
    p1 = win_1d_s.addPlot()
    vb_p2 = add_2nd_axis_to_plot(p1, label_axis_1="Intensity (a.u.)",
                                  label_axis_2="Intensity (a.u.)")
    p1.plot(ttheta, z_1_sum_e, symbol='o', pen={'color': 0.8, 'width': 2})
    p1.plot(ttheta, z_1_sum, pen=pg_pen_black)
    p1.addItem(err_s)
    p2 = pg.PlotCurveItem(ttheta, z_1_sum_e-z_1_sum, pen=pg_pen_black_thick)
    vb_p2.addItem(p2)

    win_1d_m = pg.GraphicsLayoutWidget()
    p3 = win_1d_m.addPlot()
    vb_p4 = add_2nd_axis_to_plot(p3, label_axis_1="Intensity (a.u.)",
                                  label_axis_2="Intensity (a.u.)")
    p3.plot(ttheta, z_1_diff_e, symbol='o', pen={'color': 0.8, 'width': 2})
    p3.plot(ttheta, z_1_diff, pen=pg_pen_black)
    p3.addItem(err_m)
    p4 = pg.PlotCurveItem(ttheta, z_1_diff_e-z_1_diff, pen=pg_pen_black_thick)
    vb_p4.addItem(p4)

    [x_range_12, y_range_12] = p1.viewRange()
    y_range = y_range_12[1]-y_range_12[0]
    vb_p2.setRange(yRange=(-0.15*y_range, 0.85*y_range))
    [x_range_34, y_range_34] = p3.viewRange()
    y_range = y_range_34[1]-y_range_34[0]
    p3.setYRange(-0.50*y_range, 0.50*y_range)
    vb_p4.setRange(yRange=(-0.15*y_range, 0.85*y_range))

    yrange_2 = vb_p2.viewRange()[1]
    ysize_2 = 0.05*(yrange_2[1]-yrange_2[0])
    yshift_2 = 0.5*0.5*0.5*ysize_2*len(l_pd2d_peak)
    yrange_4 = vb_p4.viewRange()[1]
    ysize_4 = 0.05*(yrange_4[1]-yrange_4[0])
    yshift_4 = 0.5*ysize_4*len(l_pd2d_peak)
    for pd2d_peak in l_pd2d_peak:
        numpy_ttheta = numpy.array(pd2d_peak.ttheta, dtype=float)
        np_zero = 0.*numpy_ttheta
        ind_miller_2 = pg.ErrorBarItem(
            x=numpy_ttheta, y=np_zero+yshift_2, top=np_zero+yshift_2+ysize_2,
            bottom=np_zero-yshift_2+ysize_2, beam=0.0)
        ind_miller_4 = pg.ErrorBarItem(
            x=numpy_ttheta, y=np_zero+yshift_4, top=np_zero+yshift_4+ysize_4,
            bottom=np_zero-yshift_4+ysize_4, beam=0.0)
        vb_p2.addItem(ind_miller_2)
        vb_p4.addItem(ind_miller_4)
        yshift_2 -= 0.5*ysize_2
        yshift_4 -= 1.0*ysize_4

    cb_s = QtWidgets.QPushButton("2D sum model")
    cb_s.clicked.connect(lambda: matrix_widget(
        win, z_sum, x_step, y_step, x_point_0, y_point_0))

    cb_s_e = QtWidgets.QPushButton("2D sum exp")
    cb_s_e.clicked.connect(lambda: matrix_widget(
        win, z_sum_e_0, x_step, y_step, x_point_0, y_point_0))

    cb_m = QtWidgets.QPushButton("2D diff model")
    cb_m.clicked.connect(lambda: matrix_widget(
        win, z_diff, x_step, y_step, x_point_0, y_point_0))

    cb_m_e = QtWidgets.QPushButton("2D diff exp")
    cb_m_e.clicked.connect(lambda: matrix_widget(
        win, z_diff_e_0, x_step, y_step, x_point_0, y_point_0))

    matrix_widget(win, z_diff, x_step, y_step, x_point_0, y_point_0)

    layout = QtWidgets.QGridLayout()
    layout.addWidget(cb_s, 0, 0)
    layout.addWidget(cb_s_e, 0, 1)
    layout.addWidget(cb_m, 0, 2)
    layout.addWidget(cb_m_e, 0, 3)
    layout.addWidget(win, 1, 0, 1, 4)
    widget = QtWidgets.QWidget()
    widget.setLayout(layout)
    # win = matrix_widget(data, x_step, y_step, x_point_0, y_point_0)

    # docks.append(pg_da.Dock("2D measurements: sum", widget=imv_s))

    docks.append(pg_da.Dock("Projection of sum", widget=win_1d_s,
                            autoOrientation=False))
    docks.append(pg_da.Dock("Projection of difference", widget=win_1d_m,
                            autoOrientation=False))
    docks.append(pg_da.Dock("Plot", widget=widget, autoOrientation=False))
    return tuple(docks)


def dock_diffrn(obj: Diffrn, thread: QtCore.QThread):
    """Docks for Diffrn objects."""
    docks = []
    f_setup = obj.setup is not None
    f_diffrn_radiation = obj.diffrn_radiation is not None
    f_diffrn_orient_matrix = obj.diffrn_orient_matrix is not None
    f_diffrn_refln = obj.diffrn_refln is not None
    f_phase = obj.phase is not None
    if not(f_setup & f_diffrn_radiation & f_diffrn_orient_matrix &
           f_diffrn_refln & f_phase):
        layout_actions = QtWidgets.QHBoxLayout()
        if not f_setup:
            cb_1 = QtWidgets.QPushButton("Add setup")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            layout_actions.addWidget(cb_1)
        if not f_diffrn_radiation:
            cb_1 = QtWidgets.QPushButton("Add diffrn_radiation")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [DiffrnRadiation()],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if not f_diffrn_orient_matrix:
            cb_1 = QtWidgets.QPushButton("Add diffrn_orient_matrix")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(
                obj, [DiffrnOrientMatrix(ub_11=1., ub_12=0., ub_13=0.,
                                         ub_21=0., ub_22=1., ub_23=0.,
                                         ub_31=0., ub_32=0., ub_33=1.,)],
                thread))
            layout_actions.addWidget(cb_1)
        if not f_diffrn_refln:
            cb_1 = QtWidgets.QPushButton("Add diffrn_refln")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [DiffrnReflnL()],
                                                   thread))
            layout_actions.addWidget(cb_1)
        if not f_phase:
            cb_1 = QtWidgets.QPushButton("Add phase")
            cb_1.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
            cb_1.setMaximumHeight(ICON_SIZE)
            cb_1.setMinimumHeight(ICON_SIZE)
            cb_1.clicked.connect(lambda: add_items(obj, [Phase(label="phase")],
                                                   thread))
            layout_actions.addWidget(cb_1)
        layout_actions.addStretch(1)
        widget_actions = QtWidgets.QWidget()
        widget_actions.setLayout(layout_actions)
        docks.append(pg_da.Dock("Actions diffrn", widget=widget_actions,
                                autoOrientation=False))

    if f_diffrn_refln:
        diffrn_refln = obj.diffrn_refln
        x = numpy.array(diffrn_refln.fr_calc, dtype=float)
        y = numpy.array(diffrn_refln.fr, dtype=float)
        err = numpy.array(diffrn_refln.fr_sigma, dtype=float)

        if not(numpy.all(numpy.isnan(x))):
            vb_1 = pg.ViewBox()
            vb_1.setMouseMode(vb_1.RectMode)
            widget = pg.PlotWidget(viewBox=vb_1)
            p1 = widget.plotItem
            err_s = pg.ErrorBarItem(x=x, y=y, top=err,
                                    bottom=err, beam=0.01)
            # err_m = pg.ErrorBarItem(x=np_x_1, y=np_y_m_1, top=np_y_sm_1,
            #                 bottom=np_y_sm_1, beam=0.5)
            p1.scatterPlot(x=x, y=y, name="fr", title="Flip ratio",
                           symbol='o', pen={'color': 0.8, 'width': 2})
            p1.plot(x=y, y=y)
            p1.addItem(err_s)
            docks.append(pg_da.Dock("Diffrn", widget=widget,
                                    autoOrientation=False))

            s1 = diffrn_refln.report_agreement_factor_exp()
            diffrn_orient_matrix = obj.diffrn_orient_matrix
            s2 = ""
            if f_diffrn_orient_matrix:
                s2 = diffrn_refln.report_chi_sq_exp(diffrn_orient_matrix.cell)
            area = QtWidgets.QScrollArea()
            area.setWidgetResizable(True)
            widget = QtWidgets.QLabel()
            widget.setText(s1+2*"\n"+s2)
            area.setWidget(widget)
            docks.append(pg_da.Dock("Difrn Refln", widget=area,
                                    autoOrientation=False))

    if f_diffrn_orient_matrix:
        diffrn_orient_matrix = obj.diffrn_orient_matrix
        widget = QtWidgets.QLabel()
        widget.setText(str(diffrn_orient_matrix.u))
        docks.append(pg_da.Dock("Diffrn Orient Matrix", widget=widget,
                                autoOrientation=False))

    f_refine_ls = obj.refine_ls is not None
    if f_refine_ls:
        docks.extend(list(dock_refine_ls(obj.refine_ls)))

    return tuple(docks)



"""WMethods class."""
from PyQt5 import QtWidgets, QtGui
from types import FunctionType


class WMethods(QtWidgets.QScrollArea):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WMethods, self).__init__(parent)
        self.object = None

        self.setWidgetResizable(True)

        self.wlist = QtWidgets.QListWidget()
        self.wlist.clicked.connect(self.do_func)

        self.setWidget(self.wlist)

    def set_thread(self, thread):
        """Set text."""
        self.mythread = thread

    def set_wfunction(self, wfunction):
        """Set text."""
        self.wfunction = wfunction

    def get_methods(self, obj):
        """Get methods."""
        qfont = QtGui.QFont()
        qfont.setBold(True)
        self.wlist.clear()
        self.object = obj
        l_method = [_1 for _1, _2 in type(obj).__dict__.items()
                    if ((type(_2) == FunctionType) &
                        (not(_1.startswith("_"))))]
        for method in l_method:
            func = getattr(obj, method)
            l_param = [_ for _ in
                       func.__code__.co_varnames[:func.__code__.co_argcount]
                       if _ != "self"]
            s_par = ""
            if len(l_param) > 0:
                s_par = ", ".join(l_param)
            s_val = f"{method:}({s_par:})"

            list_widget_item = QtWidgets.QListWidgetItem(s_val)
            list_widget_item.setFont(qfont)
            # list_widget_item.setToolTip(get_help_for_attribute(obj, method))
            self.wlist.addItem(list_widget_item)
        self.wlist.setSortingEnabled(True)

    def do_func(self, func_name):
        """Do func."""
        func_name = self.wlist.currentItem().text().split("(")[0]
        func = getattr(self.object, func_name)
        # type(self.object).__dict__[func_name]
        self.wfunction.set_function(func, self.mythread)


def do_obj_func(obj, func_name, wfunction, thread):
    """Do obj function."""
    func = type(obj).__dict__[func_name]
    wfunction.set_function(func, thread)



"""WConsole class."""
from PyQt5 import QtWidgets, QtCore, QtGui


class WConsole(QtWidgets.QScrollArea):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WConsole, self).__init__(parent)

        self.setWidgetResizable(True)

        self.wlabel = QtWidgets.QLabel()
        self.wlabel.setSizePolicy(
            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding))
        self.wlabel.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
        self.wlabel.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.wlabel.setAlignment(QtCore.Qt.AlignTop)
        self.wlabel.setWordWrap(True)
        self.wlabel.setStyleSheet("background-color:white;")

        self.setWidget(self.wlabel)

    def setText(self, text: str):
        """Set text."""
        self.wlabel.setText(text)



"""WFunction class."""
from PyQt5 import QtCore
import pyqtgraph.dockarea as pg_da

from cryspy_editor.widgets.w_presentation import w_for_presentation


class WDockArea(pg_da.DockArea):
    """
    WDockArea class.

    Attributes
    ----------
        - thread
    Mehtods
    -------
        - set_object
    """

    def __init__(self, parent=None):
        super(WDockArea, self).__init__(parent)
        self.thread = None

    def set_thread(self, thread: QtCore.QThread):
        """Set thread."""
        self.thread = thread

    def normal_clear(self):
        self.clear()
        if self.count() != 0:
            self.normal_clear()
        

    def set_object(self, obj):
        """Set text."""
        if obj is None:
            return
        self.normal_clear()
        # When number of docks is high the one clear method is not enough
        # I think that it is the bug of pyqtgraph.dockarea.
        # if self.count() != 0:
        #     self.clear()

        docks = []
        docks = w_for_presentation(obj, self.thread)

        l_ordered_docks = []
        i_actions = 0
        for i_dock, dock in enumerate(docks):
            dock.setParent(self)
            if dock.title().startswith("Actions"):
                dock.label.hide()
                l_ordered_docks.insert(i_actions, dock)
                i_actions += 1
            else:
                l_ordered_docks.append(dock)

        for i_dock, dock in enumerate(l_ordered_docks):
            if (i_dock < i_actions):
                dock.setStretch(x=1, y=1)
                if i_dock == 0:
                    self.addDock(dock, "bottom")
                else:
                    self.addDock(dock, "right", l_ordered_docks[i_dock-1])
                    
            elif (i_dock == i_actions):
                dock.setStretch(x=10, y=10)
                self.addDock(dock, "bottom")
            elif (i_dock == i_actions+1):
                dock.setStretch(x=10, y=10)
                self.addDock(dock, "right", l_ordered_docks[i_dock-1])
            elif (i_dock == i_actions+2):
                dock.setStretch(x=1, y=1)
                self.addDock(dock, "bottom")
            else:
                dock.setStretch(x=1, y=1)
                self.addDock(dock, "right", l_ordered_docks[i_dock-1])
